__all__ = ["behaviour_knowledge_space_combiner",
           "borda_count_combiner",
           "combiner",
           "complementary_output_combiner",
           "cosine_similarity_combiner",
           "decision_templates_combiner",
           "dempster_shafer_combiner",
           "k_nearest_neighbors_combiner",
           "macro_majority_vote_combiner",
           "micro_majority_vote_combiner",
           "naive_bayes_combiner",
           "neural_network_combiner",
           "simple_average_combiner",
           "weighted_voting_combiner"]